from ._ConfigsBase import ConfigsBase

class MappingConfigBase(ConfigsBase):
    def __init__(self, comment_char: str, assignment_char: str, writable: bool = True, filepath: str = None):
        self.__comment_char = comment_char
        self.__assignment_char = assignment_char
        super().__init__(writable = writable, filepath = filepath)

    @property
    def comment(self) -> str:
        return self.__comment_char

    @property
    def assignment(self) -> str:
        return self.__assignment_char

    def read(self, filepath: str):
        with open(filepath, "r") as file:
            for line in file.readlines():
                line = line.rstrip("\n").strip()
                if line != "" and line[0] != self.__comment_char:
                    self.add_item(*line.split(self.__comment_char, 1)[0].split(self.__assignment_char))

def make_custom_mapping(comment_char: str, assignment_char: str):
    class CustomMappingConfig(MappingConfigBase):
        def __init__(self, writable: bool = True, filepath: str = None):
            super().__init__(comment_char = comment_char,
                            assignment_char = assignment_char,
                            writable = writable,
                            filepath = filepath)

        @staticmethod
        def from_file(filepath: str, writable: bool = False):
            return CustomMappingConfig(writable, filepath)
    
    return CustomMappingConfig

class PropertiesConfig(MappingConfigBase):
    def __init__(self, writable: bool = True, filepath: str = None):
        super().__init__(comment_char = "#",
                         assignment_char = "=",
                         writable = writable,
                         filepath = filepath)

    @staticmethod
    def from_file(filepath: str, writable: bool = False):
        return PropertiesConfig(writable, filepath)



from enum import Enum

# Class to help prevent circular reference. See EOF for import
class PlayingCard(object):
    class AllowedValues(Enum):
        pass
    class AllowedSuits(Enum):
        pass

class IPlayingCard(object):
    """
    A card from a standard deck of 52 playing cards
    """

    @property
    def Value(self) -> PlayingCard.AllowedValues:
        """
        The card's value
        """
        raise NotImplementedError()

    @property
    def Suit(self) -> PlayingCard.AllowedSuits:
        """
        The card's suit
        """
        raise NotImplementedError()


from collections.abc import Callable, Sequence
from functools import wraps
from typing import cast as typing_cast, Any, TypeVar

from matplotlib import pyplot as plt, _preprocess_data
from matplotlib.axes import Axes
from matplotlib.collections import PolyCollection
from matplotlib.colors import Colormap
from matplotlib.typing import ColorType
from matplotlib._docstring import dedent_interpd
import numpy as np

from .._global_settings import settings_object as Settings
from ..MPI import mpi_sum, mpi_mean, mpi_gather_array



T = TypeVar("T")

class HexbinRenderer(object):
    """
    Automate the creation of matplotlib hexbin plots using custom bin value calculations.

    MPI support available, however bin value and alpha functions are expected to handle MPI
    communication internally before returning the same bin value accross all ranks! Built-in
    methods have support for this.
    """

    def __init__(
        self,
        bin_statistic:     Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]|None = None,
        bin_value_min:     float|None          = None,
        bin_value_max:     float|None          = None,
        bin_value_default: float               = -np.inf,
        cmap:              str|Colormap|None   = "viridis",
        gridsize:          int|tuple[int, int] = 100,
        edgecolor:         ColorType|None      = None,
    ) -> None:
        
        self.__hexes: PolyCollection|None = None

        self.__bin_statistic = bin_statistic if bin_statistic is not None else HexbinRenderer.bin_statistic_count
        self.__create_bin_function = HexbinRenderer._create_hexbin_colour_function(self.__bin_statistic)
        self.__bin_min = bin_value_min
        self.__bin_max = bin_value_max
        self.__bin_default = bin_value_default

        self.__cmap = cmap
        self.__gridsize = gridsize
        self.__edgecolor = edgecolor

        self.__alpha_global:                 float|None = None
        self.__alpha_values:                 np.ndarray[tuple[int], np.dtype[np.float16]]|None = None
        self.__alpha_calculator:             Callable[[float, float, float, float], Callable[[float, np.ndarray[tuple[int], np.dtype[np.int64]]], float]]|None = None
        self.__alpha_calculator_min_return:  float|None = None
        self.__alpha_calculator_max_return:  float|None = None
        self.__alpha_calculator_min:         float|None = None
        self.__alpha_calculator_cap:         float|None = None
        self.__alpha_calculator_use_unbound: float|None = None

    @property
    def hexes(self) -> PolyCollection|None:
        """
        `PolyCollection|None` Result of calling matplotlib hexbin method.

        This is only set after the renderer has been plotted for the first time.
        """
        return self.__hexes

    @property
    def bin_value_min(self) -> float|None:
        """
        `float|None` Configured minimum value a bin may be assigned.
        """
        return self.__bin_min
    @bin_value_min.setter
    def bin_value_min(self, value: float|None) -> None:
        """
        `float|None` Configured minimum value a bin may be assigned.
        """
        self.__bin_min = value

    @property
    def bin_value_max(self) -> float|None:
        """
        `float|None` Configured maximum value a bin may be assigned.
        """
        return self.__bin_max
    @bin_value_max.setter
    def bin_value_max(self, value: float|None) -> None:
        """
        `float|None` Configured maximum value a bin may be assigned.
        """
        self.__bin_max = value

    @property
    def bin_value_default(self) -> float:
        """
        `float` Configured default value a bin is assigned when empty.
        """
        return self.__bin_default
    @bin_value_default.setter
    def bin_value_default(self, value: float) -> None:
        """
        `float` Configured default value a bin is assigned when empty.
        """
        self.__bin_default = value

    @property
    def bin_alpha_floor(self) -> float:
        """
        `float` Configured minimum alpha a bin may be assigned. Below this, bins are transparent
        (alpha = 0).

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        return typing_cast(float, self.__alpha_calculator_min)
    @bin_alpha_floor.setter
    def bin_alpha_floor(self, value: float) -> None:
        """
        `float` Configured minimum alpha a bin may be assigned. Below this, bins are transparent
        (alpha = 0).

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        self.__alpha_calculator_min = value

    @property
    def bin_alpha_cap(self) -> float:
        """
        `float` Configured maximum alpha a bin may be assigned. Above this, bins are capped at this
        value.

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        return typing_cast(float, self.__alpha_calculator_cap)
    @bin_alpha_cap.setter
    def bin_alpha_cap(self, value: float) -> None:
        """
        `float` Configured maximum alpha a bin may be assigned. Above this, bins are capped at this
        value.

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        self.__alpha_calculator_cap = value

    @property
    def bin_alpha_value_min(self) -> float:
        """
        `float` Configured return value from the alpha calculation function that should be mapped
        to an alpha of 0 (transparent).

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        return typing_cast(float, self.__alpha_calculator_min_return)
    @bin_alpha_value_min.setter
    def bin_alpha_value_min(self, value: float) -> None:
        """
        `float` Configured return value from the alpha calculation function that should be mapped
        to an alpha of 0 (transparent).

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        self.__alpha_calculator_min_return = value

    @property
    def bin_alpha_value_max(self) -> float:
        """
        `float` Configured return value from the alpha calculation function that should be mapped
        to an alpha of 1 (opaque).

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        return typing_cast(float, self.__alpha_calculator_max_return)
    @bin_alpha_value_max.setter
    def bin_alpha_value_max(self, value: float) -> None:
        """
        `float` Configured return value from the alpha calculation function that should be mapped
        to an alpha of 1 (opaque).

        Raises a `NotImplementedError` if alpha is not configured to be dynamic.
        """
        if self.__alpha_calculator is None:
            raise NotImplementedError("Alpha is not dynamic. Call `set_alpha_dynamic` first.")
        self.__alpha_calculator_max_return = value

    def set_alpha_all_bins(self, alpha: float) -> None:
        """
        Set the alpha of all hexes.

        Parameters:
            `float|None` alpha:
                The alpha value to use in the range [0 -> 1].
                Set to `None` to clear any alpha related configuration.

        See also:
            `set_alpha_per_bin`
            `set_alpha_dynamic`
        """
        if alpha is not None and (alpha < 0 or alpha > 1):
            raise ValueError("Alpha value outside the range [0 -> 1].")
        self.__alpha_global                 = alpha
        self.__alpha_values                 = None
        self.__alpha_calculator             = None
        self.__alpha_calculator_min_return  = None
        self.__alpha_calculator_max_return  = None
        self.__alpha_calculator_min         = None
        self.__alpha_calculator_cap         = None
        self.__alpha_calculator_use_unbound = None

    def set_alpha_per_bin(self, alpha_values: np.ndarray[tuple[int], np.dtype[np.float16]]) -> None:
        """
        Set the alpha of each hex individually.

        Parameters:
            `float|None` alpha:
                The alpha value to use in the range [0 -> 1].
                Set to `None` to clear any alpha related configuration.

        See also:
            `set_alpha_per_bin`
            `set_alpha_dynamic`
        """
        is_invalid = (alpha_values < 0) | (alpha_values > 1)
        if is_invalid.any():
            raise ValueError("One or more alpha value(s) outside the range [0 -> 1].")
        self.__alpha_global                 = None
        self.__alpha_values                 = alpha_values.astype(np.float16)
        self.__alpha_calculator             = None
        self.__alpha_calculator_min_return  = None
        self.__alpha_calculator_max_return  = None
        self.__alpha_calculator_min         = None
        self.__alpha_calculator_cap         = None
        self.__alpha_calculator_use_unbound = None

    def set_alpha_dynamic(
        self,
        alpha_function:                       Callable[[float, np.ndarray[tuple[int], np.dtype[np.int64]]], float],
        alpha_function_min:                   float = 0.0,
        alpha_function_max:                   float = 1.0,
        alpha_min:                            float = 0.0,
        alpha_max:                            float = 1.0,
        calculation_uses_unbounded_bin_value: bool  = False
    ) -> None:
        """
        Set the alpha of an individual hex according to its value.

        Values of alpha need to be a float in the range [0 -> 1]. While the return value(s) of
        `alpha_function` may return values outside of this range, the following parameters should
        be used to control how the results of `alpha_function` are mapped to this range:

            `alpha_function_min` and `alpha_function_max` maps the range of return values between 0
            and 1 (inclusive) using the minimum and maximum values respectively.

            `alpha_min` limits what the minimum alpha value is that should actually be displayed.

            `alpha_max` caps the maximum alpha value with larger values capped at this value.

        When using MPI, alpha calculation functions that rely on the array of indexes MUST
        implement their own MPI support!

        Parameters:
            `Callable[[float, numpy.ndarray[(I,), numpy.int64]], float]` alpha_function:
                Function that calculates the value of alpha for a single bin.
                Takes two parameters: the computed value of the bin and the indexes of data
                elements that fall within the bin. Returns a single float value.
                Returned values ought to be in the range [0 -> 1] unless `alpha_function_min` and
                `alpha_function_max` are set.
            (optional) `float` alpha_function_min:
                Minimum return value that should be mapped to an alpha of 0 (invisible).
                Default is `0.0`.
            (optional) `float` alpha_function_max:
                Maximum return value that should be mapped to an alpha of 1 (opaque).
                Default is `1.0`.
            (optional) `float` alpha_min:
                Minimum alpha value to display. Values below this are invisible (alpha = 0).
                Must be in the range [0 -> 1) and less than `alpha_max`.
                Default is `0.0`.
            (optional) `float` alpha_max:
                Maximum alpha value to display. Values above this are capped at this value.
                Must be in the range (0 -> 1] and greater than `alpha_min`.
                Default is `1.0`.
            (optional) `bool` calculation_uses_unbounded_bin_value:
                Should the calculation of alpha use the calculated bin value before or after
                enforcement of any limits.
                Default is `False`.

        See also:
            `set_alpha_all_bins`
            `set_alpha_per_bin`
        """

        if alpha_min < 0:
            raise ValueError(f"Argument for \"alpha_min\" ({alpha_min}) less than 0.")
        if alpha_max > 1:
            raise ValueError(f"Argument for \"alpha_max\" ({alpha_max}) greater than 1.")
        if alpha_max <= alpha_min:
            raise ValueError(f"Argument for \"alpha_min\" ({alpha_min}) greater than argument for \"alpha_max\" ({alpha_max}).")

        self.__alpha_global                 = None
        self.__alpha_values                 = None
        self.__alpha_calculator             = HexbinRenderer._create_hexbin_alpha_function(alpha_function)
        self.__alpha_calculator_min_return  = alpha_function_min
        self.__alpha_calculator_max_return  = alpha_function_max
        self.__alpha_calculator_min         = alpha_min
        self.__alpha_calculator_cap         = alpha_max
        self.__alpha_calculator_use_unbound = calculation_uses_unbounded_bin_value

    def plot(
        self,
        x: np.ndarray[tuple[int], np.dtype[Any]],
        y: np.ndarray[tuple[int], np.dtype[Any]],
        axis: Axes|None = None,
        **kwargs
    ) -> None:
        """
        Plot a hexbin using the colour scheme.

        kwargs will be passed to the `plt.hexbin` function call (or that of the provided axis).
        """

        update_axis: bool = axis is None
        if update_axis:
            axis = plt.gca()

        alpha_calculator = None
        if self.__alpha_calculator is not None:
            alpha_calculator = self.__alpha_calculator(
                typing_cast(float, self.__alpha_calculator_min_return),
                typing_cast(float, self.__alpha_calculator_max_return),
                typing_cast(float, self.__alpha_calculator_min),
                typing_cast(float, self.__alpha_calculator_cap)
            )

        self.__hexes = HexbinRenderer.__modified_hexbin(
            x                 = x,
            y                 = y,
            C                 = np.arange(len(x)),
            reduce_C_function = self.__create_bin_function(
                self.__bin_min,
                self.__bin_max,
                self.__bin_default,
                alpha_calculator,
                self.__alpha_calculator_use_unbound
            ),
            alpha             = self.__alpha_global,
            gridsize          = self.__gridsize,
            cmap              = self.__cmap,
            vmin              = self.__bin_min,
            vmax              = self.__bin_max,
            edgecolor         = self.__edgecolor,
            mincnt            = 0, # Important to always call the colour reduction function
            **kwargs
        )

        if update_axis:
            plt.sci(typing_cast(PolyCollection, self.__hexes))

        if self.__alpha_values is not None:
            typing_cast(PolyCollection, self.__hexes).set_alpha(typing_cast(Sequence[float], self.__alpha_values))

    @staticmethod
    def _create_hexbin_colour_function(statistic: Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]):
        """
        Creates a function around the bin statistic function to work with the HexbinRenderer's
        `plot` method. This is required to allow wrapping of the statistic method for applying the
        configured bin limits.

        This is for internal use and does not need to be manually applied to input functions. This
        is handled internally.
        """

        def ready_hexbin_colour_function(min_value: float|None = None, max_value: float|None = None, default_bin_value: float = -np.inf, alpha_function: Callable[[float, np.ndarray[tuple[int], np.dtype[np.int64]]], float]|None = None, alpha_calculator_use_unbound: bool|None = None):

            @wraps(statistic)
            def calculate_bin_colour_and_alpha(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> tuple[float, float]:
                total_indexes: int = mpi_sum([len(indices)])
                if total_indexes == 0:
                    return default_bin_value, alpha_function(default_bin_value, indices) if alpha_function is not None else 1
                try:
                    result = statistic(indices)
                except Exception as e:
                    if Settings.debug and Settings.verbose:
                        raise e
                    return default_bin_value, alpha_function(default_bin_value, np.array([], dtype = np.int64)) if alpha_function is not None else 1
                if min_value is not None and result < min_value:
                    return min_value, alpha_function(result if alpha_calculator_use_unbound else min_value, indices) if alpha_function is not None else 1
                elif max_value is not None and result > max_value:
                    return max_value, alpha_function(result if alpha_calculator_use_unbound else max_value, indices) if alpha_function is not None else 1
                else:
                    return result, alpha_function(result, indices) if alpha_function is not None else 1

            return calculate_bin_colour_and_alpha

        return ready_hexbin_colour_function

    @staticmethod
    def _create_hexbin_alpha_function(alpha_function: Callable[[float, np.ndarray[tuple[int], np.dtype[np.int64]]], float]):
        """
        Creates a function around the bin alpha function to work with the HexbinRenderer's
        `plot` method. This is required to allow wrapping of the function for applying the
        configured limits.

        This is for internal use and does not need to be manually applied to input functions. This
        is handled internally.
        """

        def ready_hexbin_alpha_function(min_alpha: float, max_alpha: float, alpha_floor: float, alpha_cap: float):

            @wraps(alpha_function)
            def calculate_bin_alpha(bin_value: float, indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
                result = (alpha_function(bin_value, indices) - min_alpha) / (max_alpha - min_alpha)
                if result < alpha_floor:
                    return 0
                elif result > alpha_cap:
                    return alpha_cap
                else:
                    return result

            return calculate_bin_alpha

        return ready_hexbin_alpha_function

    # Built-in bin statistic functions

    @staticmethod
    def bin_statistic_count(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
        """
        Assign hexbin value to the number of elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(I,), numpy.int64]` indices:
                The indexes into the x and y data arrays that identify the points that fall into
                the bin.

        Returns `float`:
            The number of elements within the bin.
        """
        return float(mpi_sum(len(indices)))

    @staticmethod
    def bin_statistic_log10_count(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
        """
        Assign hexbin value to log_10 of the number of elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(I,), numpy.int64]` indices:
                The indexes into the x and y data arrays that identify the points that fall into
                the bin.

        Returns `float`:
            The number of elements within the bin.
        """
        return np.log10(HexbinRenderer.bin_statistic_count(indices = indices))

    @staticmethod
    def gather_bin_values(data: np.ndarray[tuple[int], np.dtype[T]], indices: np.ndarray[tuple[int], np.dtype[np.int64]]) -> np.ndarray[tuple[int], np.dtype[T]]:
        """
        Helper function for gathering bin data separated across multiple MPI ranks.
        This is in effect an 'allgather' operation.
        """
        return mpi_gather_array(data[indices], allgather = True)

    @staticmethod
    def create_bin_statistic_sum(data: np.ndarray[tuple[int], np.dtype[float]]) -> Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Assign hexbin value to the number of elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(N,), float]` data:
                Data elements - the same shape as the x and y data arrays.

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            A bin statistic for the number of elements within the bin.
        """
        def bin_statistic_sum(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
            Assign hexbin value to the number of elements that fall within each bin.

            Parameters:
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                The number of elements within the bin.
            """
            return float(mpi_sum(data[indices]))
        return bin_statistic_sum

    @staticmethod
    def create_bin_statistic_log10_sum(data: np.ndarray[tuple[int], np.dtype[float]], initial_offset: float = 0.0, final_offset: float = 1.0) -> Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Assign hexbin value to log_10 of the number of elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(N,), float]` data:
                Data elements - the same shape as the x and y data arrays.
            (optional) `float` initial_offset:
                Value to add to each data value BEFORE summation. This is necessary when the data
                values can be 0!
                Default is 0.
            (optional) `float` final_offset:
                A value to add to the final result - useful for unit conversions.
                NOTE: this value is not passed through the log_10 function. When applying unit
                conversions, most units will require a value of `-numpy.log10(value_of_unit)`!
                Default is 1.

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            A bin statistic for log_10 of the number of elements within the bin.
        """
        def bin_statistic_log10_sum(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
            Assign hexbin value to log_10 of the number of elements that fall within each bin.

            Parameters:
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                The number log_10 of elements within the bin.
            """
            return float(np.log10(mpi_sum(data[indices] + initial_offset))) + final_offset
        return bin_statistic_log10_sum

    @staticmethod
    def create_bin_statistic_mean(data: np.ndarray[tuple[int], np.dtype[float]], weights: np.ndarray[tuple[int], np.dtype[float]]|None = None) -> Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Assign hexbin value to the mean of elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(N,), float]` data:
                Data elements - the same shape as the x and y data arrays.
            (optional) `numpy.ndarray[(N,), float]|None` weights:
                One weight per data element.
                Default is `None`.

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            A bin statistic for the mean of elements within the bin.
        """
        def bin_statistic_weighted_mean(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
            Assign hexbin value to the mean of elements that fall within each bin.

            Parameters:
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                The mean of elements within the bin.
            """
            return float(mpi_mean(data[indices], weights = weights[indices] if weights is not None else None))
        return bin_statistic_weighted_mean

    @staticmethod
    def create_bin_statistic_median(data: np.ndarray[tuple[int], np.dtype[float]]) -> Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Assign hexbin value to the median of elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(N,), float]` data:
                Data elements - the same shape as the x and y data arrays.

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            A bin statistic for the median of elements within the bin.
        """
        def bin_statistic_median(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
        Assign hexbin value to the median of elements that fall within each bin.

            Parameters:
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                The median of elements within the bin.
            """

            return float(np.median(HexbinRenderer.gather_bin_values(data, indices)))
        return bin_statistic_median

    @staticmethod
    def create_bin_statistic_percentile(data: np.ndarray[tuple[int], np.dtype[float]], percentile: float) -> Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Assign hexbin value to the percentile value from elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(N,), float]` data:
                Data elements - the same shape as the x and y data arrays.
            `float` percentile:
                The percentile at which compute the value.
                [0 -> 100]

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            A bin statistic for the percentile value of elements within the bin.
        """
        def bin_statistic_percentile(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
            Assign hexbin value to the percentile value from elements that fall within each bin.

            Parameters:
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                The percentile value of elements within the bin.
            """

            return float(np.percentile(HexbinRenderer.gather_bin_values(data, indices), percentile))
        return bin_statistic_percentile
    
    @staticmethod
    def create_bin_statistic_log10_mean(data: np.ndarray[tuple[int], np.dtype[float]], weights: np.ndarray[tuple[int], np.dtype[float]]|None = None, initial_offset: float = 0.0, final_offset: float = 1.0) -> Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Assign hexbin value to the mean of elements that fall within each bin.

        Parameters:
            `numpy.ndarray[(N,), float]` data:
                Data elements - the same shape as the x and y data arrays.
            (optional) `numpy.ndarray[(N,), float]|None` weights:
                One weight per data element.
                Default is `None`.
            (optional) `float` initial_offset:
                Value to add to each data value BEFORE summation. This is necessary when the data
                values can be 0!
                Default is 0.
            (optional) `float` final_offset:
                A value to add to the final result - useful for unit conversions.
                NOTE: this value is not passed through the log_10 function. When applying unit
                conversions, most units will require a value of `-numpy.log10(value_of_unit)`!
                Default is 1.

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            A bin statistic for log_10 of the mean of elements within the bin.
        """
        def bin_statistic_log10_weighted_mean(indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
            Assign hexbin value to the mean of elements that fall within each bin.

            Parameters:
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                log_10 of the mean of elements within the bin.
            """
            return float(np.log10(mpi_mean(data[indices] + initial_offset, weights = weights[indices] if weights is not None else None))) + final_offset
        return bin_statistic_log10_weighted_mean

    # Built-in bin alpha functions

    @staticmethod
    def bin_alpha_value(bin_value: float, indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
        """
        Use the value of the bin to set the alpha value of the hex.

        Parameters:
            `float` bin_value:
                The calculated value for the bin.
            `numpy.ndarray[(I,), numpy.int64]` indices:
                The indexes into the x and y data arrays that identify the points that fall into
                the bin.

        Returns `float`:
            The value of the bin.
        """
        return bin_value

    @staticmethod
    def bin_alpha_reciprocal_value(bin_value: float, indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
        """
        Use the reciprocal of the value of the bin to set the alpha value of the hex.

        Parameters:
            `float` bin_value:
                The calculated value for the bin.
            `numpy.ndarray[(I,), numpy.int64]` indices:
                The indexes into the x and y data arrays that identify the points that fall into
                the bin.

        Returns `float`:
            The reciprocal (1/x) of the value of the bin.
            If `len(indices)` is 0, numpy.Infinity is returned.
        """
        return (1 / bin_value) if bin_value != 0 else np.Infinity

    @staticmethod
    def create_bin_alpha_statistic(statistic: Callable[[np.ndarray[tuple[int], np.dtype[np.int64]]], float]) -> Callable[[float, np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Use an existing statistic function to set the alpha value of the hex.

        Parameters:
            `Callable[[np.ndarray[(I,), numpy.int64]], float]` statistic:
                A bin statistic function used to calculate a value from the indexes giving which
                elements fall within a given bin.

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            The wrapped statistic function.
        """
        def bin_statistic_sum(bin_value: float, indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
            Use an existing statistic function to set the alpha value of the hex.

            Parameters:
                `float` bin_value:
                    The calculated value for the bin.
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                The statistic for the bin.
            """
            return statistic(indices)
        return bin_statistic_sum

    @staticmethod
    def create_bin_alpha_value_converter(conversion_function: Callable[[float], float]) -> Callable[[float, np.ndarray[tuple[int], np.dtype[np.int64]]], float]:
        """
        Use a function to set the alpha value of the hex using the hex's value.

        Parameters:
            `Callable[[float], float]` conversion_function:
                Function used to convert a bin's calculated value into an alpha value.

        Returns `Callable[[numpy.ndarray[(I,), numpy.int64]], float]`:
            The wrapped conversion function.
        """
        def bin_statistic_sum(bin_value: float, indices: np.ndarray[tuple[int], np.dtype[np.int64]], /) -> float:
            """
            Use a function to set the alpha value of the hex using the hex's value.

            Parameters:
                `float` bin_value:
                    The calculated value for the bin.
                `numpy.ndarray[(I,), numpy.int64]` indices:
                    The indexes into the x and y data arrays that identify the points that fall
                    into the bin.

            Returns `float`:
                The alpha value for the bin.
            """
            return conversion_function(bin_value)
        return bin_statistic_sum



    @staticmethod
    @_preprocess_data(replace_names=["x", "y", "C"], label_namer="y")
    @dedent_interpd
    def __modified_hexbin(self, x, y, C=None, gridsize=100, bins=None,
               xscale='linear', yscale='linear', extent=None,
               cmap=None, norm=None, vmin=None, vmax=None,
               alpha=None, linewidths=None, edgecolors='face',
               reduce_C_function=np.mean, mincnt=None, marginals=False,
               **kwargs) -> PolyCollection:
        """
        Lifted from matplotlib.axes._axes as there is no modular version.
        Edits are made to facilitate the on-the-fly calculation of alpha values.
        MPI support has also been added to allow use with large datasets.
        Source code is only edited where necessary and is otherwise left as-is.

        The original docstring is as follows:

        Make a 2D hexagonal binning plot of points *x*, *y*.

        If *C* is *None*, the value of the hexagon is determined by the number
        of points in the hexagon. Otherwise, *C* specifies values at the
        coordinate (x[i], y[i]). For each hexagon, these values are reduced
        using *reduce_C_function*.

        Parameters
        ----------
        x, y : array-like
            The data positions. *x* and *y* must be of the same length.

        C : array-like, optional
            If given, these values are accumulated in the bins. Otherwise,
            every point has a value of 1. Must be of the same length as *x*
            and *y*.

        gridsize : int or (int, int), default: 100
            If a single int, the number of hexagons in the *x*-direction.
            The number of hexagons in the *y*-direction is chosen such that
            the hexagons are approximately regular.

            Alternatively, if a tuple (*nx*, *ny*), the number of hexagons
            in the *x*-direction and the *y*-direction. In the
            *y*-direction, counting is done along vertically aligned
            hexagons, not along the zig-zag chains of hexagons; see the
            following illustration.

            .. plot::

               import numpy
               import matplotlib.pyplot as plt

               np.random.seed(19680801)
               n= 300
               x = np.random.standard_normal(n)
               y = np.random.standard_normal(n)

               fig, ax = plt.subplots(figsize=(4, 4))
               h = ax.hexbin(x, y, gridsize=(5, 3))
               hx, hy = h.get_offsets().T
               ax.plot(hx[24::3], hy[24::3], 'ro-')
               ax.plot(hx[-3:], hy[-3:], 'ro-')
               ax.set_title('gridsize=(5, 3)')
               ax.axis('off')

            To get approximately regular hexagons, choose
            :math:`n_x = \\sqrt{3}\\,n_y`.

        bins : 'log' or int or sequence, default: None
            Discretization of the hexagon values.

            - If *None*, no binning is applied; the color of each hexagon
              directly corresponds to its count value.
            - If 'log', use a logarithmic scale for the colormap.
              Internally, :math:`log_{10}(i+1)` is used to determine the
              hexagon color. This is equivalent to ``norm=LogNorm()``.
            - If an integer, divide the counts in the specified number
              of bins, and color the hexagons accordingly.
            - If a sequence of values, the values of the lower bound of
              the bins to be used.

        xscale : {'linear', 'log'}, default: 'linear'
            Use a linear or log10 scale on the horizontal axis.

        yscale : {'linear', 'log'}, default: 'linear'
            Use a linear or log10 scale on the vertical axis.

        mincnt : int >= 0, default: *None*
            If not *None*, only display cells with at least *mincnt*
            number of points in the cell.

        marginals : bool, default: *False*
            If marginals is *True*, plot the marginal density as
            colormapped rectangles along the bottom of the x-axis and
            left of the y-axis.

        extent : 4-tuple of float, default: *None*
            The limits of the bins (xmin, xmax, ymin, ymax).
            The default assigns the limits based on
            *gridsize*, *x*, *y*, *xscale* and *yscale*.

            If *xscale* or *yscale* is set to 'log', the limits are
            expected to be the exponent for a power of 10. E.g. for
            x-limits of 1 and 50 in 'linear' scale and y-limits
            of 10 and 1000 in 'log' scale, enter (1, 50, 1, 3).

        Returns
        -------
        `~matplotlib.collections.PolyCollection`
            A `.PolyCollection` defining the hexagonal bins.

            - `.PolyCollection.get_offsets` contains a Mx2 array containing
              the x, y positions of the M hexagon centers.
            - `.PolyCollection.get_array` contains the values of the M
              hexagons.

            If *marginals* is *True*, horizontal
            bar and vertical bar (both PolyCollections) will be attached
            to the return collection as attributes *hbar* and *vbar*.

        Other Parameters
        ----------------
        %(cmap_doc)s

        %(norm_doc)s

        %(vmin_vmax_doc)s

        alpha : float between 0 and 1, optional
            The alpha blending value, between 0 (transparent) and 1 (opaque).

        linewidths : float, default: *None*
            If *None*, defaults to :rc:`patch.linewidth`.

        edgecolors : {'face', 'none', *None*} or color, default: 'face'
            The color of the hexagon edges. Possible values are:

            - 'face': Draw the edges in the same color as the fill color.
            - 'none': No edges are drawn. This can sometimes lead to unsightly
              unpainted pixels between the hexagons.
            - *None*: Draw outlines in the default color.
            - An explicit color.

        reduce_C_function : callable, default: `numpy.mean`
            The function to aggregate *C* within the bins. It is ignored if
            *C* is not given. This must have the signature::

                def reduce_C_function(C: array) -> float

            Commonly used functions are:

            - `numpy.mean`: average of the points
            - `numpy.sum`: integral of the point values
            - `numpy.amax`: value taken from the largest point

            By default will only reduce cells with at least 1 point because some
            reduction functions (such as `numpy.amax`) will error/warn with empty
            input. Changing *mincnt* will adjust the cutoff, and if set to 0 will
            pass empty input to the reduction function.

        data : indexable object, optional
            DATA_PARAMETER_PLACEHOLDER

        **kwargs : `~matplotlib.collections.PolyCollection` properties
            All other keyword arguments are passed on to `.PolyCollection`:

            %(PolyCollection:kwdoc)s

        See Also
        --------
        hist2d : 2D histogram rectangular bins
        """
        self._process_unit_info([("x", x), ("y", y)], kwargs, convert=False)

        x, y, C = cbook.delete_masked_points(x, y, C)

        # Set the size of the hexagon grid
        if np.iterable(gridsize):
            nx, ny = gridsize
        else:
            nx = gridsize
            ny = int(nx / math.sqrt(3))
        # Count the number of data in each hexagon
        x = np.asarray(x, float)
        y = np.asarray(y, float)

        # Will be log()'d if necessary, and then rescaled.
        tx = x
        ty = y

        if xscale == 'log':
            if np.any(x <= 0.0):
                raise ValueError(
                    "x contains non-positive values, so cannot be log-scaled")
            tx = np.log10(tx)
        if yscale == 'log':
            if np.any(y <= 0.0):
                raise ValueError(
                    "y contains non-positive values, so cannot be log-scaled")
            ty = np.log10(ty)
        if extent is not None:
            xmin, xmax, ymin, ymax = extent
        else:
            xmin, xmax = (tx.min(), tx.max()) if len(x) else (0, 1)
            ymin, ymax = (ty.min(), ty.max()) if len(y) else (0, 1)

            # to avoid issues with singular data, expand the min/max pairs
            xmin, xmax = mtransforms.nonsingular(xmin, xmax, expander=0.1)
            ymin, ymax = mtransforms.nonsingular(ymin, ymax, expander=0.1)

        nx1 = nx + 1
        ny1 = ny + 1
        nx2 = nx
        ny2 = ny
        n = nx1 * ny1 + nx2 * ny2

        # In the x-direction, the hexagons exactly cover the region from
        # xmin to xmax. Need some padding to avoid roundoff errors.
        padding = 1.e-9 * (xmax - xmin)
        xmin -= padding
        xmax += padding
        sx = (xmax - xmin) / nx
        sy = (ymax - ymin) / ny
        # Positions in hexagon index coordinates.
        ix = (tx - xmin) / sx
        iy = (ty - ymin) / sy
        ix1 = np.round(ix).astype(int)
        iy1 = np.round(iy).astype(int)
        ix2 = np.floor(ix).astype(int)
        iy2 = np.floor(iy).astype(int)
        # flat indices, plus one so that out-of-range points go to position 0.
        i1 = np.where((0 <= ix1) & (ix1 < nx1) & (0 <= iy1) & (iy1 < ny1),
                      ix1 * ny1 + iy1 + 1, 0)
        i2 = np.where((0 <= ix2) & (ix2 < nx2) & (0 <= iy2) & (iy2 < ny2),
                      ix2 * ny2 + iy2 + 1, 0)

        d1 = (ix - ix1) ** 2 + 3.0 * (iy - iy1) ** 2
        d2 = (ix - ix2 - 0.5) ** 2 + 3.0 * (iy - iy2 - 0.5) ** 2
        bdist = (d1 < d2)

        if C is None:  # [1:] drops out-of-range points.
            counts1 = np.bincount(i1[bdist], minlength=1 + nx1 * ny1)[1:]
            counts2 = np.bincount(i2[~bdist], minlength=1 + nx2 * ny2)[1:]
            accum = np.concatenate([counts1, counts2]).astype(float)
            if mincnt is not None:
                accum[accum < mincnt] = np.nan
            C = np.ones(len(x))
        else:
            # store the C values in a list per hexagon index
            Cs_at_i1 = [[] for _ in range(1 + nx1 * ny1)]
            Cs_at_i2 = [[] for _ in range(1 + nx2 * ny2)]
            for i in range(len(x)):
                if bdist[i]:
                    Cs_at_i1[i1[i]].append(C[i])
                else:
                    Cs_at_i2[i2[i]].append(C[i])
            if mincnt is None:
                mincnt = 1
            accum = np.array(
                [reduce_C_function(acc) if len(acc) >= mincnt else np.nan
                 for Cs_at_i in [Cs_at_i1, Cs_at_i2]
                 for acc in Cs_at_i[1:]],  # [1:] drops out-of-range points.
                float)

        good_idxs = ~np.isnan(accum)

        offsets = np.zeros((n, 2), float)
        offsets[:nx1 * ny1, 0] = np.repeat(np.arange(nx1), ny1)
        offsets[:nx1 * ny1, 1] = np.tile(np.arange(ny1), nx1)
        offsets[nx1 * ny1:, 0] = np.repeat(np.arange(nx2) + 0.5, ny2)
        offsets[nx1 * ny1:, 1] = np.tile(np.arange(ny2), nx2) + 0.5
        offsets[:, 0] *= sx
        offsets[:, 1] *= sy
        offsets[:, 0] += xmin
        offsets[:, 1] += ymin
        # remove accumulation bins with no data
        offsets = offsets[good_idxs, :]
        accum = accum[good_idxs]

        polygon = [sx, sy / 3] * np.array(
            [[.5, -.5], [.5, .5], [0., 1.], [-.5, .5], [-.5, -.5], [0., -1.]])

        if linewidths is None:
            linewidths = [mpl.rcParams['patch.linewidth']]

        if xscale == 'log' or yscale == 'log':
            polygons = np.expand_dims(polygon, 0) + np.expand_dims(offsets, 1)
            if xscale == 'log':
                polygons[:, :, 0] = 10.0 ** polygons[:, :, 0]
                xmin = 10.0 ** xmin
                xmax = 10.0 ** xmax
                self.set_xscale(xscale)
            if yscale == 'log':
                polygons[:, :, 1] = 10.0 ** polygons[:, :, 1]
                ymin = 10.0 ** ymin
                ymax = 10.0 ** ymax
                self.set_yscale(yscale)
            collection = mcoll.PolyCollection(
                polygons,
                edgecolors=edgecolors,
                linewidths=linewidths,
                )
        else:
            collection = mcoll.PolyCollection(
                [polygon],
                edgecolors=edgecolors,
                linewidths=linewidths,
                offsets=offsets,
                offset_transform=mtransforms.AffineDeltaTransform(
                    self.transData),
            )

        # Set normalizer if bins is 'log'
        if cbook._str_equal(bins, 'log'):
            if norm is not None:
                _api.warn_external("Only one of 'bins' and 'norm' arguments "
                                   f"can be supplied, ignoring {bins=}")
            else:
                norm = mcolors.LogNorm(vmin=vmin, vmax=vmax)
                vmin = vmax = None
            bins = None

        # autoscale the norm with current accum values if it hasn't been set
        if norm is not None:
            if norm.vmin is None and norm.vmax is None:
                norm.autoscale(accum)

        if bins is not None:
            if not np.iterable(bins):
                minimum, maximum = min(accum), max(accum)
                bins -= 1  # one less edge than bins
                bins = minimum + (maximum - minimum) * np.arange(bins) / bins
            bins = np.sort(bins)
            accum = bins.searchsorted(accum)

        collection.set_array(accum)
        collection.set_cmap(cmap)
        collection.set_norm(norm)
        collection.set_alpha(alpha)
        collection._internal_update(kwargs)
        collection._scale_norm(norm, vmin, vmax)

        corners = ((xmin, ymin), (xmax, ymax))
        self.update_datalim(corners)
        self._request_autoscale_view(tight=True)

        # add the collection last
        self.add_collection(collection, autolim=False)
        if not marginals:
            return collection

        # Process marginals
        bars = []
        for zname, z, zmin, zmax, zscale, nbins in [
                ("x", x, xmin, xmax, xscale, nx),
                ("y", y, ymin, ymax, yscale, 2 * ny),
        ]:

            if zscale == "log":
                bin_edges = np.geomspace(zmin, zmax, nbins + 1)
            else:
                bin_edges = np.linspace(zmin, zmax, nbins + 1)

            verts = np.empty((nbins, 4, 2))
            verts[:, 0, 0] = verts[:, 1, 0] = bin_edges[:-1]
            verts[:, 2, 0] = verts[:, 3, 0] = bin_edges[1:]
            verts[:, 0, 1] = verts[:, 3, 1] = .00
            verts[:, 1, 1] = verts[:, 2, 1] = .05
            if zname == "y":
                verts = verts[:, :, ::-1]  # Swap x and y.

            # Sort z-values into bins defined by bin_edges.
            bin_idxs = np.searchsorted(bin_edges, z) - 1
            values = np.empty(nbins)
            for i in range(nbins):
                # Get C-values for each bin, and compute bin value with
                # reduce_C_function.
                ci = C[bin_idxs == i]
                values[i] = reduce_C_function(ci) if len(ci) > 0 else np.nan

            mask = ~np.isnan(values)
            verts = verts[mask]
            values = values[mask]

            trans = getattr(self, f"get_{zname}axis_transform")(which="grid")
            bar = mcoll.PolyCollection(
                verts, transform=trans, edgecolors="face")
            bar.set_array(values)
            bar.set_cmap(cmap)
            bar.set_norm(norm)
            bar.set_alpha(alpha)
            bar._internal_update(kwargs)
            bars.append(self.add_collection(bar, autolim=False))

        collection.hbar, collection.vbar = bars

        def on_changed(collection):
            collection.hbar.set_cmap(collection.get_cmap())
            collection.hbar.set_cmap(collection.get_cmap())
            collection.vbar.set_clim(collection.get_clim())
            collection.vbar.set_clim(collection.get_clim())

        collection.callbacks.connect('changed', on_changed)

        return collection



from typing import Any, Iterable
from typing import TypeVar

from ._Cacheable import Cacheable

T = TypeVar("T", bound = Cacheable)

class CacheableList[T](list[T], Cacheable):
    """
    A list of cacheable objects.
    
    This class is used to handle a list of cacheable objects, allowing for the 
    serialization and deserialization of the list as a whole.
    
    Attributes:
        list[Cacheable] _cacheable_list:
            The list of cacheable objects.
    """
    
    def __init__(self, iterable: Iterable[T]) -> None:
        super().__init__(iterable)

    @classmethod
    def __from_cache_data__(cls, data: dict[str, Any]) -> "CacheableList[T]":
        """
        Load the list of cacheable objects from a cache target.
        
        Parameters:
            dict[str, Any] data:
                The data to load the list of cacheable objects from.
        
        Returns:
            CacheableList -> The loaded list of cacheable objects.
        """
        number_of_items: int = data["count"]
        return cls([data["types"][i].__from_cache_data__(data["data"]) for i in range(number_of_items)])

    def __get_cache_data__(self) -> dict[str, Any]:
        return {
            "count": len(self),
            "types": [type(item) for item in self],
            "data": [item.__get_cache_data__() for item in self],
        }



import os
import pickle
from typing import Any, Type, TypeVar

from ._Cacheable import Cacheable

T = TypeVar("T", bound = Cacheable)

class CacheTarget(object):
    """
    A class to manage caching of data to and from pickle files.

    Parameters:
        str relative_file_path:
            The relative file path to the cache file.
            This path is relative to the root directory provided when saving or loading data.
    """

    def __init__(self, relative_file_path: str) -> None:
        self.__relative_file_path: str = relative_file_path

    @property
    def relative_filepath(self) -> str:
        """
        The relative file path to the cache file.

        Returns:
            str -> The relative file path as a string.
        """
        return self.__relative_file_path
    
    def get_filepath(self, root_directory: str) -> str:
        """
        Constructs a full file path by combining the provided root directory 
        with the relative file path of the cache target.

        Parameters:
            str root_directory:
                The root directory to which the relative file path will be appended.

        Returns:
            str -> The full file path constructed from the root directory and 
                 the relative file path.
        """
        return os.path.join(root_directory, self.__relative_file_path)

    def check_exists(self, root_directory: str) -> bool:
        """
        Check if the cache file exists at the specified location.

        Parameters:
            str root_directory:
                The root directory to which the relative file path will be appended.

        Returns:
            bool -> True if the cache file exists, False otherwise.
        """
        return os.path.exists(self.get_filepath(root_directory))
    
    def save_data(self, root_directory: str, data: dict[str, Any]) -> None:
        """
        Save data to a pickle file using this target from a given root directory.

        Parameters:
            str root_directory:
                The root directory to which the relative file path will be appended.
            dict[str, Any] data:
                The data to save to the pickle file.
        """
        filepath = self.get_filepath(os.path.abspath(root_directory))
        directory = os.path.dirname(self.get_filepath(root_directory))
        if not os.path.exists(directory):
            paths_to_make = []
            test_path = directory
            while not os.path.exists(test_path):
                paths_to_make.append(test_path)
                test_path = os.path.dirname(test_path)
            for path in paths_to_make[::-1]:
                os.mkdir(path)
        with open(filepath, "wb") as file:
            pickle.dump(data, file)

    def load_data(self, root_directory: str) -> dict[str, Any]:
        """
        Load data from a pickle file using this target from a given root directory.

        Parameters:
            str root_directory:
                The root directory to which the relative file path will be appended.

        Returns:
            dict[str, Any] -> The data loaded from the pickle file.
        """
        filepath = self.get_filepath(os.path.abspath(root_directory))
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"Unable to locate cache file at \"{filepath}\".")
        with open(filepath, "rb") as file:
            return pickle.load(file)
        
    # Handle objects of type Cacheable

    def save_object(self, root_directory: str, obj: Cacheable) -> None:
        """
        Save a Cacheable object to a pickle file using this target from a given root directory.

        Parameters:
            str root_directory:
                The root directory to which the relative file path will be appended.
            object obj:
                The Cacheable object to save.
        """
        if not isinstance(obj, Cacheable):
            raise TypeError("The object must be an instance of Cacheable.")
        self.save_data(root_directory, obj.__get_cache_data__())

    def load_object(self, root_directory: str, cls: Type[T]) -> T:
        """
        Load a Cacheable object from a pickle file using this target from a given root directory.

        Parameters:
            str root_directory:
                The root directory to which the relative file path will be appended.
            Type[T] cls:
                The class type of the Cacheable object to load.

        Returns:
            T -> The Cacheable object loaded from the pickle file.
        """
        data = self.load_data(root_directory)
        return cls.__from_cache_data__(data)



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.patches import Polygon

def polygon_y_error_region(x_coordinates, y_upper_coordinates, y_lower_coordinates, axis = None, **plotting_kwargs):

    poly = Polygon(np.concatenate((np.array([[x_coordinates[i], y_lower_coordinates[i]] for i in range(len(x_coordinates))], dtype = float),
                                   np.array([[x_coordinates[i], y_upper_coordinates[i]] for i in range(len(x_coordinates) - 1, -1, -1)], dtype = float)),
                                  axis = 0),
                   closed = False,
                   **plotting_kwargs)
    
    if axis is not None:
        if axis == "default":
            plt.gca().add_patch(poly)
        else:
            axis.add_patch(poly)

    return poly

def polygon_x_error_region(x_left_coordinates, x_right_coordinates, y_coordinates, axis = None, **plotting_kwargs):

    poly = Polygon(np.concatenate((np.array([[x_left_coordinates[i], y_coordinates[i]] for i in range(len(y_coordinates))], dtype = float),
                                   np.array([[x_right_coordinates[i], y_coordinates[i]] for i in range(len(y_coordinates) - 1, -1, -1)], dtype = float)),
                                  axis = 0),
                   closed = False,
                   **plotting_kwargs)
    
    if axis is not None:
        if axis == "default":
            plt.gca().add_patch(poly)
        else:
            axis.add_patch(poly)

    return poly



from ._VersionInformation import VersionInfomation
from ._PersonInfomation import PersonInfomation
from ._AuthorInfomation import AuthorInfomation
from ._Rect import IRect, GenericRect, Rect, DiscreteRect



from ._array_visuliser import ArrayVisuliser
from ._autoproperty import AutoProperty, TypedAutoProperty, NullableTypedAutoProperty, TypeCastAutoProperty
from ._async import start_main_async
from . import _directorys_and_imports as DirectoryTools
from ._multiItterator import MultiItterator
from ._script_wrapper import ScriptWrapper, ScriptWrapper_ParamSpec_Error, ScriptWrapper_Param_ValidityError, \
                             ScriptWrapper_Param_ValueError, ScriptWrapper_Param_TypeError, ScriptWrapper_Param_UnsetError, \
                             ScriptWrapper_Param_ConflictError, ScriptWrapper_Param_MissingRequirementError
from ._Struct import Struct, CacheableStruct
from ._CacheableFunction import CacheableFunction, PolynomialFunction, LinearFunction, GaussianFunction
from ._typeshield import TypeShield_Base, TypeShield, NestedTypeShield
from ._cast import Cast, non_nullable_cast, nullable_cast, NestedCast
from ._common_casts import *
from . import StringLiterals
from . import Validators
from ._edit_locals import update_locals, read_locals, use_locals
from ._stopwatch import IStopwatch, SimpleStopwatch
from ._ScreenResolution import ScreenResolution



class MultiItterator(object):
    """
    Allows for itteration of multiple collections similtaniously.
    Produces a tuple with the first element being the index used and the other
    items being the corisponding elements from the collections provided.

    Arguments:
        params[collection] collections -> Collections of the same length to itterate through.
    """

    def __init__(self, *collections):
        self.__items = tuple(collections)
        self.__length = len(self.__items[0])
        self.__nextIndex = 0

        sameLength = True
        for i in range(1, len(self.__items)):
            if len(self.__items[i]) != self.__length:
                sameLength = False
                break

            if not sameLength:
                raise ValueError("The colections provided were not all of the same length.")

    def __len__(self):
        return self.__length

    def __iter__(self):
        self.__nextIndex = 0
        return self

    def __next__(self):
        """
        Returns:
            tuple -> (index, item1[index], item2[index], ...)
        """
        if self.__nextIndex < self.__length:
            index = self.__nextIndex
            self.__nextIndex += 1
            return (index, *[collection[index] for collection in self.__items])
        else:
            raise StopIteration


from ._ISpinner import ISpinner
import random

class Spinner(ISpinner):
    """
    Creates spinner objects with n sides and labels. Labels are selected and returned by calling the avalable methods.

    __init__():
        list labels --> list of labels - one for each side
        int seed --> seed for the random generator (deafult is None)
    """
    def __init__(self, labels: list, seed: int = None):
        self.__Generator: random.Random = random.Random() if seed == None else random.Random(seed)
        self.Labels: list = list(labels)
        self.Sides: int = len(labels)

    def ContextSpin(self):
        """
        Randomly selects a side and retuns the side and the corisponding label.
        """
        index = self.__Generator.randint(0, self.Sides)
        return (index, self.Labels[index])

    def Spin(self):
        """
        Randomly selects a side and retuns the corisponding label.
        """
        return self.Labels[self.__Generator.randint(0, self.Sides)]
        
    def IndexSpin(self):
        """
        Randomly selects a side and retuns the side number.
        """
        return self.__Generator.randint(0, self.Sides)


from ._console import Console
from ._console import OldConsole# Depreciated!!!
from ._stopwatch import Stopwatch

print_info = Console.print_info
print_verbose_info = Console.print_verbose_info
print_warning = Console.print_warning
print_verbose_warning = Console.print_verbose_warning
print_error = Console.print_error
print_verbose_error = Console.print_verbose_error
print_debug = Console.print_debug
print_verbose_debug = Console.print_verbose_debug



from ._console import Console
from ...Tools._stopwatch import IStopwatch, SimpleStopwatch
from ..._global_settings import settings_object as _settings_object

from collections.abc import Sequence

import numpy as np
import datetime



class Stopwatch(IStopwatch):
    """
    Feature rich stopwatch implementation with printing options.
    """

    @staticmethod
    def start_new(name: str|None = None) -> "Stopwatch":
        stopwatch = Stopwatch(name)
        stopwatch.start()
        return stopwatch
    @staticmethod
    def synchronise_start_new(synchronise: IStopwatch, name: str|None = None) -> "Stopwatch":
        stopwatch = Stopwatch()
        stopwatch.synchronise_start(synchronise)
        return stopwatch
    
    @staticmethod
    def new_group(number: int, names: Sequence[str|None]|None = None) -> "tuple[Stopwatch]":
        return tuple([Stopwatch(names[i] if names is not None else None) for i in range(number)])
    @staticmethod
    def start_new_group(number: int, names: Sequence[str|None]|None = None) -> "tuple[Stopwatch]":
        stopwatches = tuple([Stopwatch(names[i] if names is not None else None) for i in range(number)])
        stopwatches[0].start()
        for i in range(1, number):
            stopwatches[i].synchronise_start(stopwatches[0])
        return stopwatches
    @staticmethod
    def synchronise_start_new_group(number: int, synchronise: "IStopwatch", names: Sequence[str|None]|None = None) -> "tuple[Stopwatch]":
        return tuple([Stopwatch.synchronise_start_new(synchronise, names[i] if names is not None else None) for i in range(number)])

    def __init__(self, name: str|None = None):
        self.__name = name if name is not None else "Stopwatch"
        self.__internal_stopwatch = SimpleStopwatch()

    #TODO: add context manager methods for easy time reporting (retain as variable to use time data, otherwise set to print time)

    @staticmethod
    def __strftime(dt: datetime.timedelta|float):
        if isinstance(dt, float):
            dt = datetime.timedelta(seconds = dt)
        return (datetime.datetime(year = 1, month = 1, day = 1) + dt).strftime(_settings_object.time_format_precise)
    
    def reset(self) -> None:
        self.__internal_stopwatch.reset()

    def start(self) -> None:
        self.__internal_stopwatch.start()

    def print_start(self) -> None:
        self.start()
        Console.custom_print("|TIME|", f"{self.__name}: START {Stopwatch.__strftime(self.get_start_time())}", show_insert = True)

    def synchronise_start(self, synchronise: IStopwatch) -> None:
        self.__internal_stopwatch.synchronise_start(synchronise)

    def lap(self) -> float:
        return self.__internal_stopwatch.lap()

    def print_lap(self) -> float:
        dt = self.lap()
        Console.custom_print("|TIME|", f"{self.__name}: LAP {Stopwatch.__strftime(dt)} ({Stopwatch.__strftime(self.get_elapsed_time())})", show_insert = True)
        return dt

    def stop(self) -> float:
        return self.__internal_stopwatch.stop()

    def print_stop(self) -> float:
        dt = self.stop()
        Console.custom_print("|TIME|", f"{self.__name}: LAP {Stopwatch.__strftime(dt)} ({Stopwatch.__strftime(self.get_elapsed_time())}) STOPPED", show_insert = True)
        return dt
    
    def __enter__(self) -> "Stopwatch":
        self.start()
        Console.custom_print("|TIME|", f"{self.__name}: STARTED @ {Stopwatch.__strftime(self.start_time)}", show_insert = True)
        return self

    def __exit__(self, type, value, tb) -> None:
        self.stop()
        Console.custom_print("|TIME|", f"{self.__name}: STOPPED @ {Stopwatch.__strftime(self.stop_time)}", show_insert = True)
        Console.custom_print("|TIME|", f"{self.__name}: ELAPSED   {Stopwatch.__strftime(self.get_elapsed_time())}", show_insert = True)
    
    def get_start_time(self) -> float:
        return self.__internal_stopwatch.get_start_time()
    
    def get_elapsed_time(self) -> float:
        return self.__internal_stopwatch.get_elapsed_time()
    
    def get_elapsed_time_lap(self) -> float:
        return self.__internal_stopwatch.get_elapsed_time_lap()

    def get_elapsed_time_last_lap(self) -> float:
        return self.__internal_stopwatch.get_elapsed_time_last_lap()

    def __str__(self) -> str:
        return self.__internal_stopwatch.__str__()

    def __repr__(self) -> str:
        if self.running:
            return f"Stopwatch: {self.get_elapsed_time()}"
        elif self.stopped:
            return f"Stopwatch: {self.get_elapsed_time()} (stopped)"
        else:
            return "Stopwatch: not yet started"
        
    def __float__(self):
        return self.__internal_stopwatch.__float__()
        
    def __int__(self):
        return self.__internal_stopwatch.__int__()

    @property
    def running(self) -> bool:
        return self.__internal_stopwatch.running
    @property
    def started(self) -> bool:
        return self.__internal_stopwatch.started
    @property
    def stopped(self) -> bool:
        return self.__internal_stopwatch.stopped

    @property
    def start_time(self) -> float|None:
        return self.__internal_stopwatch.start_time
    @property
    def lap_times(self) -> tuple[float]:
        return self.__internal_stopwatch.lap_times
    @property
    def stop_time(self) -> float|None:
        return self.__internal_stopwatch.stop_time

    def __create_print_sting__elapsed_time(self) -> str:
        return f"{self.__name}: {Stopwatch.__strftime(self.get_elapsed_time())}"
    def print_elapsed_time(self, show_insert: bool = True, **kwargs) -> None:
        Console.custom_print("|TIME|", self.__create_print_sting__elapsed_time(), show_insert = show_insert, **kwargs)
    def print_verbose_elapsed_time(self, show_insert: bool = True, verbosity_level: int = -1, **kwargs) -> None:
        Console.custom_print_verbose("|TIME|", self.__create_print_sting__elapsed_time(), show_insert = show_insert, verbosity_level = verbosity_level, **kwargs)
    def print_verbose_debug_elapsed_time(self, show_insert: bool = True, verbosity_level: int = -1, **kwargs) -> None:
        Console.custom_print_verbose_debug("|TIME|", self.__create_print_sting__elapsed_time(), show_insert = show_insert, verbosity_level = verbosity_level, **kwargs)
    def print_debug_elapsed_time(self, show_insert: bool = True, **kwargs) -> None:
        Console.custom_print_debug("|TIME|", self.__create_print_sting__elapsed_time(), show_insert = show_insert, **kwargs)

    def __create_print_sting__elapsed_time_lap(self) -> str:
        return f"{self.__name}: LAP {Stopwatch.__strftime(self.get_elapsed_time_lap())}"
    def print_elapsed_time_lap(self, show_insert: bool = True, **kwargs) -> None:
        Console.custom_print("|TIME|", self.__create_print_sting__elapsed_time_lap(), show_insert = show_insert, **kwargs)
    def print_verbose_elapsed_time_lap(self, show_insert: bool = True, verbosity_level: int = -1, **kwargs) -> None:
        Console.custom_print_verbose("|TIME|", self.__create_print_sting__elapsed_time_lap(), show_insert = show_insert, verbosity_level = verbosity_level, **kwargs)
    def print_verbose_debug_elapsed_time_lap(self, show_insert: bool = True, verbosity_level: int = -1, **kwargs) -> None:
        Console.custom_print_verbose_debug("<TIME>", self.__create_print_sting__elapsed_time_lap(), show_insert = show_insert, verbosity_level = verbosity_level, **kwargs)
    def print_debug_elapsed_time_lap(self, show_insert: bool = True, **kwargs) -> None:
        Console.custom_print_debug("<TIME>", self.__create_print_sting__elapsed_time_lap(), show_insert = show_insert, **kwargs)

    def __create_print_sting__elapsed_time_last_lap(self) -> str:
        return f"{self.__name}: LAP(-1) {Stopwatch.__strftime(self.get_elapsed_time_last_lap())}"
    def print_elapsed_time_last_lap(self, show_insert: bool = True, **kwargs) -> None:
        Console.custom_print("|TIME|", self.__create_print_sting__elapsed_time_last_lap(), show_insert = show_insert, **kwargs)
    def print_verbose_elapsed_time_last_lap(self, show_insert: bool = True, verbosity_level: int = -1, **kwargs) -> None:
        Console.custom_print_verbose("|TIME|", self.__create_print_sting__elapsed_time_last_lap(), show_insert = show_insert, verbosity_level = verbosity_level, **kwargs)
    def print_verbose_debug_elapsed_time_last_lap(self, show_insert: bool = True, verbosity_level: int = -1, **kwargs) -> None:
        Console.custom_print_verbose_debug("<TIME>", self.__create_print_sting__elapsed_time_last_lap(), show_insert = show_insert, verbosity_level = verbosity_level, **kwargs)
    def print_debug_elapsed_time_last_lap(self, show_insert: bool = True, **kwargs) -> None:
        Console.custom_print_debug("<TIME>", self.__create_print_sting__elapsed_time_last_lap(), show_insert = show_insert, **kwargs)



from ._CardGroup import CardGroup
from ._IPlayingCard import IPlayingCard

class CardStack(CardGroup):
    """
    A CardGroup stack data structure containing IPlayingCards

    NOTE: This is only a convenience structure that mirrors the CardStack from the C# library.
            The CardGroup can be used like a stack in the same way as a Python list.

    Atributes:
        str Name --> name of the collection

    __init__():
        str name --> name of the collection
        list cards --> list of IPlayingCards to initialise the collection with (deafult is blank list)
    """
    def __init__(self, name: str, cards: list = []):
        super().__init__(name, cards)

    def Push(self, item: IPlayingCard):
        self.append(item)

    def Pop(self) -> IPlayingCard:
        return self.pop()


from ..Tools._Struct import Struct
from ..Tools._autoproperty import NullableTypedAutoProperty
from ..Tools._typeshield import TypeShield, NestedTypeShield

from typing import Sequence

class PersonInfomation(Struct):
    given_name      = NullableTypedAutoProperty[str](TypeShield[str](str))
    family_name     = NullableTypedAutoProperty[str](TypeShield[str](str))
    email_addresses = NullableTypedAutoProperty[Sequence[str]](NestedTypeShield[Sequence[str]](str))
    website_url     = NullableTypedAutoProperty[str](TypeShield[str](str))
    #TODO: update the type checks for email addresses and URLs



import pytest
import sys

from QuasarCode.IO.Text import print_info, print_verbose_info, print_debug
from QuasarCode.Tools import ScriptWrapper

class TestClass_Tools(object):
    def test_ScriptWrapper(self):
        assert True
        def testing_function():
            print_info("SUCCESS")
            print_verbose_info("FAIL")
            print_debug("SUCCESS")
            assert True

        sys.argv.append("-d")
        wrapper = ScriptWrapper("filename.py", "author", "0.0.0", "edit_date", "description", [], [], [], [])
        
        wrapper.run(testing_function)



"""
~| Event Driven Python (edp) |~

Implements Event Driven programming in Python

Credits:
    Written by Christopher Rowe 04/2019
    
Version: 1.1
"""

from ._Event import Event
from ._UndersubscribedEventError import UndersubscribedEventError
from ._SubscriberNotPresentError import SubscriberNotPresentError
from ._NotEnoughArgumentsError import NotEnoughArgumentsError



import os
from typing import Any, Literal

from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import numpy as np

from ..Data._Rect import Rect
from ..Tools._ScreenResolution import ScreenResolution
from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty, AutoProperty_NonNullable
from ._CachedPlot import CachedPlot
from ._CachedPlotElements import CachedPlotColourbar

class CachedFigureGrid(CacheableStruct):
    """
    Design a matplotlib figure using a mosaic layout.
    This (and any added cacheable plots) may be cached to disk.
    """

    title = AutoProperty[str](allow_uninitialised = True)
    mosaic = AutoProperty_NonNullable[list[list[str]]]()
    rows = AutoProperty_NonNullable[int](default_value = 1)
    columns = AutoProperty_NonNullable[int](default_value = 1)
    figure_size = AutoProperty_NonNullable[tuple[int, int]](default_value = (6, 8))
    layout = AutoProperty[Literal["constrained", "compressed", "tight"]](allow_uninitialised = True)
    relative_heights = AutoProperty[list[float]](allow_uninitialised = True)
    relative_widths = AutoProperty[list[float]](allow_uninitialised = True)
    vertical_spacing = AutoProperty_NonNullable[float](default_value = 0)
    horizontal_spacing = AutoProperty_NonNullable[float](default_value = 0)
    plots = AutoProperty_NonNullable[dict[str, CachedPlot]]()
    colourbars = AutoProperty_NonNullable[dict[str, CachedPlotColourbar]]()
    resolution = AutoProperty_NonNullable[int](default_value = 100)
    resolution_for_files = AutoProperty[int](allow_uninitialised = True)

    @property
    def physical_rect(self) -> Rect:
        """
        Rect -> Dimensions of the figure in inches.
        """
        return Rect.create_from_size(
            x = 0.0,
            y = 0.0,
            width = self.figure_size[0],
            height = self.figure_size[1]
        )

    @property
    def physical_heights(self) -> list[float]:
        """
        list[float] -> Height of each row in inches.
        """
        total_height = self.figure_size[1]
        total_relative_height = sum(self.relative_heights) if self.relative_heights else self.rows
        return [total_height * (row_height / total_relative_height) for row_height in (self.relative_heights if self.relative_heights else [1] * self.rows)]
    
    @property
    def physical_widths(self) -> list[float]:
        """
        list[float] -> Width of each column in inches.
        """
        total_width = self.figure_size[0]
        total_relative_width = sum(self.relative_widths) if self.relative_widths else self.columns
        return [total_width * (column_width / total_relative_width) for column_width in (self.relative_widths if self.relative_widths else [1] * self.columns)]

    def __init__(self, **kwargs) -> None:
        self.__figure: Figure|None = None
        self.__axes: dict[str, Axes]|None = None
        self.__locked = False
        mosaic_provided = "mosaic" in kwargs

        super().__init__(
            cacheable_attributes = ("title", "mosaic", "rows", "columns", "figure_size", "layout", "relative_widths", "relative_heights", "vertical_spacing", "horizontal_spacing", "plots", "resolution", "resolution_for_files"),
            **kwargs
        )

        if mosaic_provided:
            if len(self.mosaic) < 1 or len(self.mosaic[0]) < 1:
                raise ValueError("Mosaic must have at least one row and one column.")
            if any([len(row) != len(self.mosaic[0]) for row in self.mosaic]):
                raise ValueError("All rows in the mosaic must have the same number of columns.")

        if not mosaic_provided:
            self.mosaic = [["." for _ in range(self.columns)] for _ in range(self.rows)]
        else:
            self.rows = len(self.mosaic)
            self.columns = max(len(row) for row in self.mosaic)

        self.plots = {}
        self.colourbars = {}
        
    def set_resolution_custom(self, width_pixels: int|None = None, height_pixels: int|None = None, file_only: bool = False) -> None:
        """
        Set the resolution of the final image to fit within a custom width.

        Parameters:
            int width_pixels*:
                The desired width in pixels.
            int height_pixels*:
                The desired height in pixels.
            bool file_only:
                Apply this only to a saved image and not when displaying the image.

            * Only one of these parameters may be specified at a time!
        """
        if width_pixels is None and height_pixels is None:
            raise ValueError("At least one of `width_pixels` or `height_pixels` must be provided.")
        elif width_pixels is not None and height_pixels is not None:
            raise ValueError("Only one of `width_pixels` or `height_pixels` may be provided at a time.")
        resolution: int = int(width_pixels / self.figure_size[0]) if width_pixels is not None else int(height_pixels / self.figure_size[1])
        if file_only:
            self.resolution_for_files = resolution
        else:
            self.resolution = resolution
            if self.__figure is not None:
                self.__figure.dpi = resolution
        
    def set_resolution_720p(self, file_only: bool = False) -> None:
        """
        Set the resolution of the final image to fit within a 720p display.

        Parameters:
            bool file_only:
                Apply this only to a saved image and not when displaying the image.
        """
        resolution: int
        target_ratio = ScreenResolution.get_720p().value.height / ScreenResolution.get_720p().value.width
        plot_ratio = self.figure_size[1] / self.figure_size[0]
        if plot_ratio <= target_ratio:
            # The width is the limiting factor
            resolution = int(ScreenResolution.get_720p().value.width / self.figure_size[0])
        else:
            # The height is the limiting factor
            resolution = int(ScreenResolution.get_720p().value.height / self.figure_size[1])
        if file_only:
            self.resolution_for_files = resolution
        else:
            self.resolution = resolution
            if self.__figure is not None:
                self.__figure.dpi = resolution

    def set_resolution_1080p(self, file_only: bool = False) -> None:
        """
        Set the resolution of the final image to fit within a 1080p display.

        Parameters:
            bool file_only:
                Apply this only to a saved image and not when displaying the image.
        """
        resolution: int
        target_ratio = ScreenResolution.get_1080p().value.height / ScreenResolution.get_1080p().value.width
        plot_ratio = self.figure_size[1] / self.figure_size[0]
        if plot_ratio <= target_ratio:
            # The width is the limiting factor
            resolution = int(ScreenResolution.get_1080p().value.width / self.figure_size[0])
        else:
            # The height is the limiting factor
            resolution = int(ScreenResolution.get_1080p().value.height / self.figure_size[1])
        if file_only:
            self.resolution_for_files = resolution
        else:
            self.resolution = resolution
            if self.__figure is not None:
                self.__figure.dpi = resolution

    def set_resolution_4K(self, file_only: bool = False) -> None:
        """
        Set the resolution of the final image to fit within a 4K display.

        Parameters:
            bool file_only:
                Apply this only to a saved image and not when displaying the image.
        """
        resolution: int
        target_ratio = ScreenResolution.get_4K().value.height / ScreenResolution.get_4K().value.width
        plot_ratio = self.figure_size[1] / self.figure_size[0]
        if plot_ratio <= target_ratio:
            # The width is the limiting factor
            resolution = int(ScreenResolution.get_4K().value.width / self.figure_size[0])
        else:
            # The height is the limiting factor
            resolution = int(ScreenResolution.get_4K().value.height / self.figure_size[1])
        if file_only:
            self.resolution_for_files = resolution
        else:
            self.resolution = resolution
            if self.__figure is not None:
                self.__figure.dpi = resolution

    def set_resolution_8K(self, file_only: bool = False) -> None:
        """
        Set the resolution of the final image to fit within a 8K display.

        Parameters:
            bool file_only:
                Apply this only to a saved image and not when displaying the image.
        """
        resolution: int
        target_ratio = ScreenResolution.EIGHT_K.value.height / ScreenResolution.EIGHT_K.value.width
        plot_ratio = self.figure_size[1] / self.figure_size[0]
        if plot_ratio <= target_ratio:
            # The width is the limiting factor
            resolution = int(ScreenResolution.EIGHT_K.value.width / self.figure_size[0])
        else:
            # The height is the limiting factor
            resolution = int(ScreenResolution.EIGHT_K.value.height / self.figure_size[1])
        if file_only:
            self.resolution_for_files = resolution
        else:
            self.resolution = resolution
            if self.__figure is not None:
                self.__figure.dpi = resolution

    def add_row(self, number: int = 1, insert_at_index: int = -1, height: float = 1.0, physical_height: bool = False, expand_figure: bool = False) -> None:
        """
        Add a new row of axes to the figure layout.

        Parameters:
            int number:
                Number of rows of this specification to add.
                Must be greater than or equal to 1.
                Default is  1.
            int insert_at_index:
                Index at which to insert the new rows.
                Default is -1, which means to append at the end.
            float height:
                Height of the new rows (relative unless specified as physical).
            bool physical_height:
                If True, the height is specified in inches.
                Note: this requires `expand_figure` to be True.
                Default is False.
            bool expand_figure:
                If True, the figure size will be expanded to accommodate the new rows.
                This is required when using `physical_height`.
                Default is False.
        """
        if number < 1:
            raise ValueError("Number of rows to add must be at least 1.")
        if insert_at_index < 0:
            insert_at_index = self.rows + insert_at_index
        if self.__locked:
            raise RuntimeError("Cannot add rows after the figure has been created. Call `clear_axes` first.")
        if physical_height and not expand_figure:
            raise ValueError("Cannot add rows with physical height without expanding the figure. Set `expand_figure` to True.")
        if self.relative_heights is None:
            self.relative_heights = [1.0] * self.rows
        if physical_height:
            initial_physical_height = self.figure_size[1]
            if expand_figure:
                self.figure_size = (self.figure_size[0], self.figure_size[1] + height * number)
            height *= (sum(self.relative_heights) if self.relative_heights else self.rows) / initial_physical_height
        elif expand_figure:
            self.figure_size = (self.figure_size[0], self.figure_size[1] + (height * (self.figure_size[1] / sum(self.relative_heights))) * number)
        for _ in range(number):
            self.mosaic.insert(insert_at_index, ["." for _ in range(self.columns)])
            self.relative_heights.insert(insert_at_index, height)
        self.rows += number

    def add_column(self, number: int = 1, insert_at_index: int = -1, width: float = 1.0, physical_width: bool = False, expand_figure: bool = False) -> None:
        """
        Add a new column of axes to the figure layout.

        Parameters:
            int number:
                Number of columns of this specification to add.
                Must be greater than or equal to 1.
                Default is  1.
            int insert_at_index:
                Index at which to insert the new columns.
                Default is -1, which means to append at the end.
            float width:
                Width of the new columns (relative unless specified as physical).
            bool physical_width:
                If True, the width is specified in inches.
                Note: this requires `expand_figure` to be True.
                Default is False.
            bool expand_figure:
                If True, the figure size will be expanded to accommodate the new columns.
                This is required when using `physical_width`.
                Default is False.
        """
        if number < 1:
            raise ValueError("Number of columns to add must be at least 1.")
        if insert_at_index < 0:
            insert_at_index = self.columns + insert_at_index
        if self.__locked:
            raise RuntimeError("Cannot add columns after the figure has been created. Call `clear_axes` first.")
        if physical_width and not expand_figure:
            raise ValueError("Cannot add columns with physical width without expanding the figure. Set `expand_figure` to True.")
        if self.relative_widths is None:
            self.relative_widths = [1.0] * self.columns
        if physical_width:
            initial_physical_width = self.figure_size[0]
            if expand_figure:
                self.figure_size = (self.figure_size[0] + width * number, self.figure_size[1])
            width *= (sum(self.relative_widths) if self.relative_widths else self.columns) / initial_physical_width
        elif expand_figure:
            self.figure_size = (self.figure_size[0] + (width * (self.figure_size[0] / sum(self.relative_widths))) * number, self.figure_size[1])
        for row in self.mosaic:
            for _ in range(number):
                row.insert(insert_at_index, ".")
        self.columns += number
        for _ in range(number):
            self.relative_widths.insert(insert_at_index, width)

    def resize_row(self, row: int|list[int], height: float, physical_height: bool = False, resize_figure: bool = False) -> None:
        """
        Alter the height of one or more rows.

        Parameters:
            int|list[int] row:
                Index(es) of row(s) to set the height of.
            float height:
                Height of the new rows (relative unless specified as physical).
            bool physical_height:
                If True, the height is specified in inches.
                Note: this requires `expand_figure` to be True.
                Default is False.
            bool expand_figure:
                If True, the figure size will be expanded to accommodate the new rows.
                This is required when using `physical_height`.
                Default is False.
        """
        if self.__locked:
            raise RuntimeError("Cannot resize rows after the figure has been created. Call `clear_axes` first.")
        if isinstance(row, int):
            row = [row]
        if self.relative_heights is None:
            self.relative_heights = [1.0] * self.rows
        initial_physical_figure_height = self.figure_size[1]
        initial_relative_figure_height = sum(self.relative_heights)
        new_relative_height = height * (self.figure_size[1] / initial_physical_figure_height) if physical_height else height
        old_relative_heights = []
        for i in row:
            if i >= self.rows or i < -self.rows:
                raise IndexError(f"Row index {i} is out of bounds for the number of rows ({self.rows}).")
            old_relative_heights.append(self.relative_heights[i])
            self.relative_heights[i] = new_relative_height
        if resize_figure:
            delta_physical_height = (new_relative_height * len(row) - sum(old_relative_heights)) * (initial_physical_figure_height / initial_relative_figure_height)
            self.figure_size = (self.figure_size[0], self.figure_size[1] + delta_physical_height)

    def resize_column(self, column: int|list[int], width: float, physical_width: bool = False, resize_figure: bool = False) -> None:
        """
        Alter the width of one or more columns.

        Parameters:
            int|list[int] column:
                Index(es) of column(s) to set the width of.
            float width:
                Width of the new columns (relative unless specified as physical).
            bool physical_width:
                If True, the width is specified in inches.
                Note: this requires `expand_figure` to be True.
                Default is False.
            bool expand_figure:
                If True, the figure size will be expanded to accommodate the new columns.
                This is required when using `physical_width`.
                Default is False.
        """
        if self.__locked:
            raise RuntimeError("Cannot resize columns after the figure has been created. Call `clear_axes` first.")
        if isinstance(column, int):
            column = [column]
        if self.relative_widths is None:
            self.relative_widths = [1.0] * self.columns
        initial_physical_figure_width = self.figure_size[0]
        initial_relative_figure_width = sum(self.relative_widths)
        new_relative_width = width * (self.figure_size[0] / initial_physical_figure_width) if physical_width else width
        old_relative_widths = []
        for j in column:
            if j >= self.columns or j < -self.columns:
                raise IndexError(f"Column index {j} is out of bounds for the number of columns ({self.columns}).")
            old_relative_widths.append(self.relative_widths[j])
            self.relative_widths[j] = new_relative_width
        if resize_figure:
            delta_physical_width = (new_relative_width * len(column) - sum(old_relative_widths)) * (initial_physical_figure_width / initial_relative_figure_width)
            self.figure_size = (self.figure_size[0] + delta_physical_width, self.figure_size[1])

    def assign_empty(self, row: int|list[int], column: int|list[int]) -> None:
        """
        Assign one or more grid elements to be empty space.
        This functions as a wrapper for `assign_plot` with the name ".".

        Parameters:
            int|list[int] row:
                Index(es) of row(s) to assign as empty.
            int|list[int] column:
                Index(es) of column(s) to assign as empty.
        """
        self.assign_plot(".", row, column)

    def assign_plot(self, name: str, row: int|list[int] = 0, column: int|list[int] = 0) -> None:
        """
        Assign one or more grid elements to an axis name.
        Row and column indexes should be sequential to avoid gaps.

        Parameters:
            int|list[int] row:
                Index(es) of row(s) to assign this name.
            int|list[int] column:
                Index(es) of column(s) to assign this name.
        """
        if isinstance(row, int):
            row = [row]
        if isinstance(column, int):
            column = [column]
        for i in row:
            if i >= self.rows or i < -self.rows:
                raise IndexError(f"Row index {i} is out of bounds for the number of rows ({self.rows}).")
            for j in column:
                if j >= self.columns or j < -self.columns:
                    raise IndexError(f"Column index {j} is out of bounds for the number of columns ({self.columns}).")
                self.mosaic[i][j] = name

    def set_plot(self, name: str, plot: CachedPlot) -> None:
        """
        Link an axis name in the figure layout with a CachedPlot instance for easy rendering.

        Parameters:
            str name:
                Name of the axis to link.
            CachedPlot plot:
                CachedPlot instance to insert.
        """
        if not any([name in row for row in self.mosaic]):
            raise KeyError(f"No plot with the name \"{name}\" exists in the mosaic. Add the plot first with `assign_plot`.")
        self.plots[name] = plot

    def remove_plot(self, name: str) -> None:
        """
        Remove an added CachedPlot instance.

        Parameters:
            str name:
                Name of the axis to unlink.
        """
        try:
            self.plots.pop(name)
        except KeyError: pass

    def set_colourbar(self, name: str, colourbar: CachedPlotColourbar, target_plot: str, target_element: str) -> None:
        """
        Link an axis name in the figure layout with a CachedPlotColourbar instance for easy rendering.
        Note: this circumvents the need to create a separate CachedPlot instance to host the colourbar.
        Adding colourbars into the layout mosaic is recommended as this allows greater control over the sizes of individual plots.

        Parameters:
            str name:
                Name of the axis to link.
            CachedPlotColourbar colourbar:
                CachedPlotColourbar instance to insert.
        """
        if name in self.colourbars:
            raise KeyError(f"A colourbar already exists with the name \"{name}\".")
        if not any([name in row for row in self.mosaic]):
            raise KeyError(f"No plot with the name \"{name}\" exists in the mosaic. Add the plot first with `assign_plot`.")
        if not any([target_plot in row for row in self.mosaic]):
            raise KeyError(f"No plot with the name \"{target_plot}\" exists in the mosaic. Add the plot first with `assign_plot`.")
        colourbar.target_element = target_element
        colourbar.target_plot = target_plot
        colourbar.add_to_axis = False
        self.colourbars[name] = colourbar
        
    def remove_colourbar(self, name: str) -> None:
        """
        Remove an added CachedPlotColourbar instance.

        Parameters:
            str name:
                Name of the axis to unlink.
        """
        try:
            self.colourbars.pop(name)
        except KeyError: pass

    def get_axis_rect_relative(self, name: str) -> Rect:
        """
        Get a Rect instance containing the relative position and size of a named axis.

        Parameters:
            str name:
                Name of the axis.

        Returns:
            Rect: Relative coordinates of the bottom-left corner and dimensions.
        """
        if name == ".":
            raise ValueError("Plot tag \".\" is reserved for empty space. To use this space, assign a name first.")
        row_indexes, column_indexes = np.where(np.array(self.mosaic, dtype = str) == name)
        heights = (self.relative_heights if self.relative_heights else [1] * self.rows   )
        widths  = (self.relative_widths  if self.relative_widths  else [1] * self.columns)
        return Rect.create_from_size(
            x = sum(widths[:column_indexes[0]]),
            y = sum(heights[:row_indexes[0]]),
            width  = sum(widths[column_indexes[0] : column_indexes[-1] + 1]),
            height = sum(heights[row_indexes[0] : row_indexes[-1] + 1])
        )

    def get_axis_rect_physical(self, name: str) -> Rect:
        """
        Get a Rect instance containing the position and size in inches of a named axis.

        Parameters:
            str name:
                Name of the axis.

        Returns:
            Rect: Coordinates of the bottom-left corner and dimensions.
        """
        relative = self.get_axis_rect_relative(name)
        total_relative_height = sum(self.relative_heights) if self.relative_heights else self.rows
        total_relative_width = sum(self.relative_widths) if self.relative_widths else self.columns
        return Rect.create_from_size(
            x = (relative.x / total_relative_width) * self.figure_size[0],
            y = (relative.y / total_relative_height) * self.figure_size[1],
            width = (relative.width / total_relative_width) * self.figure_size[0],
            height = (relative.height / total_relative_height) * self.figure_size[1]
        )

    def clear_axes(self) -> None:
        """
        Delete the figure and axes objects created by `make_figure_and_axes`.
        This will unlock the instance and allow further alteration of the layout.
        """
        if self.__locked:
            self.__figure = None
            self.__axes = None
            self.__locked = False

    def make_figure_and_axes(self, figure_kwargs: dict[str, Any]|None = None, mosaic_kwargs: dict[str, Any]|None = None, gridspec_kwargs: dict[str, Any]|None = None) -> tuple[Figure, dict[str, Axes]]:
        """
        Generate matplotlib figure and axes objects according to the layout specification.
        Doing so will lock the object's state and prevent further alterations to the layout.
        These may be subsequently accessed through the `figure` and `axes` properties.

        Parameters:
            dict[str, Any] figure_kwargs:
                Additional keyword arguments to pass to `plt.figure`.
                `figsize`, `layout`, and `dpi` are already set and MUST not be specified here.
            dict[str, Any] mosaic_kwargs:
                Additional keyword arguments to pass to `plt.subplot_mosaic`.
                `width_ratios` and `height_ratios` are already set and MUST not be specified here.
                `gridspec_kw` may be altered using the `gridspec_kwargs` parameter.
            dict[str, Any] gridspec_kwargs:
                Additional keyword arguments to pass to the gridspec constructor.
                `wspace` and `hspace` are already set and MUST not be specified here.

        Returns:
            tuple[Figure, dict[str, Axes]]:
                The created figure and a dictionary of axes keyed by their names.
        """
        if self.__locked:
            raise RuntimeError("Figure and axes have already been created. Cannot create them again without first calling `clear_axes`.")
        self.__figure = plt.figure(figsize = self.figure_size, layout = self.layout, **(figure_kwargs if figure_kwargs is not None else {}))
        self.__figure.dpi = self.resolution
        if self.title is not None:
            self.__figure.suptitle(self.title)
        self.__axes = self.__figure.subplot_mosaic(
            self.mosaic,
            width_ratios = self.relative_widths,
            height_ratios = self.relative_heights,
            gridspec_kw = { "wspace": self.horizontal_spacing, "hspace": self.vertical_spacing, **(gridspec_kwargs if gridspec_kwargs is not None else {}) },
            **(mosaic_kwargs if mosaic_kwargs is not None else {})
        )
        self.__locked = True
        return self.__figure, self.__axes
    
    @property
    def figure(self) -> Figure:
        """
        Figure -> The matplotlib figure object created by `make_figure_and_axes`.

        Raises:
            RuntimeError: If the figure has not been created yet.
        """
        if not self.__locked:
            raise RuntimeError("Figure has not been created yet. Call `make_figure_and_axes` first.")
        return self.__figure

    @property
    def axes(self) -> dict[str, Axes]:
        """
        Axes -> The dictionary of axes created by `make_figure_and_axes`.

        Raises:
            RuntimeError: If the figure has not been created yet.
        """
        if not self.__locked:
            raise RuntimeError("Axes have not been created yet. Call `make_figure_and_axes` first.")
        return self.__axes
    
    def get_axis(self, plot_tag: str) -> Axes:
        """
        Get the axes associated with a specific plot tag.

        Parameters:
            str plot_tag:
                The name of the plot to retrieve the axes for.

        Returns:
            Axes: The axes object associated with the specified plot tag.
        """
        if plot_tag == ".":
            raise ValueError("Plot tag \".\" is reserved for empty space. To use these axes, assign them a name first.")
        if not self.__locked:
            raise RuntimeError("Axes have not been created yet. Call `make_figure_and_axes` first.")
        if plot_tag not in self.__axes:
            raise KeyError(f"No axes found for plot tag \"{plot_tag}\".")
        return self.__axes[plot_tag]

    def render(self, forward_kwargs: dict[str, dict[str, dict[str, Any]]]|None = None, forward_colourbar_kwargs: dict[str, dict[str, dict[str, Any]]]|None = None) -> None:
        """
        Automatically render any associated CachedPlot instances onto their axes.
        If `make_figure_and_axes` has not been called, this will be done automatically. If you wish to control other elements of figure and axis creation, call that method first.

        Parameters:
            dict[str, dict[str, dict[str, Any]]] forward_kwargs:
                A dictionary of keyword arguments to pass to each plot's `render` method. Keys are the names of each axis.
            dict[str, dict[str, dict[str, Any]]] forward_colourbar_kwargs:
                A dictionary of keyword arguments to pass to each plot's `render` method for the colourbar. Keys are the names of each axis.
        """
        if not self.__locked:
            self.make_figure_and_axes()
        if forward_kwargs is None:
            forward_kwargs = {}
        if forward_colourbar_kwargs is None:
            forward_colourbar_kwargs = {}
        for plot_tag, plot in self.plots.items():
            plot.render(self.__figure, self.__axes[plot_tag], forward_kwargs = forward_kwargs.get(plot_tag, {}), forward_colourbar_kwargs = forward_colourbar_kwargs.get(plot_tag, {}))
        for plot_tag, colourbar in self.colourbars.items():
            colourbar.render(self.__figure, self.__axes[plot_tag], self.plots[colourbar.target_plot].plot_elements[colourbar.target_element]._result, **forward_colourbar_kwargs.get(plot_tag, {}))

    def save_png(self, filename: str, directory: str|None = None) -> None:
        """
        Save the figure as a PNG file with the specified resolution.
        The resolution for the file may be set differently to the display resolution using `resolution_for_files`.

        Parameters:
            str filename:
                The name of the file to save the figure as (without extension).
            str|None directory:
                The directory to save the file in. If None, the current working directory is used.
        """
        filepath = f"{filename}.png"
        if directory is not None:
            filepath = os.path.join(directory, filepath)
        if not self.__locked:
            raise RuntimeError("Figure has not been created yet. Call `make_figure_and_axes` first.")
        self.__figure.savefig(filepath, dpi = self.resolution_for_files if self.resolution_for_files is not None else self.resolution)
    def save_jpeg(self, filename: str, directory: str|None = None) -> None:
        """
        Save the figure as a JPEG file with the specified resolution.
        The resolution for the file may be set differently to the display resolution using `resolution_for_files`.

        Parameters:
            str filename:
                The name of the file to save the figure as (without extension).
            str|None directory:
                The directory to save the file in. If None, the current working directory is used.
        """
        filepath = f"{filename}.jpeg"
        if directory is not None:
            filepath = os.path.join(directory, filepath)
        if not self.__locked:
            raise RuntimeError("Figure has not been created yet. Call `make_figure_and_axes` first.")
        self.__figure.savefig(filepath, dpi = self.resolution_for_files if self.resolution_for_files is not None else self.resolution)
    def save_pdf(self, filename: str, directory: str|None = None) -> None:
        """
        Save the figure as a PDF file with the specified resolution.
        The resolution for the file may be set differently to the display resolution using `resolution_for_files`.

        Parameters:
            str filename:
                The name of the file to save the figure as (without extension).
            str|None directory:
                The directory to save the file in. If None, the current working directory is used.
        """
        filepath = f"{filename}.pdf"
        if directory is not None:
            filepath = os.path.join(directory, filepath)
        if not self.__locked:
            raise RuntimeError("Figure has not been created yet. Call `make_figure_and_axes` first.")
        self.__figure.savefig(filepath, dpi = self.resolution_for_files if self.resolution_for_files is not None else self.resolution)
    def save_svg(self, filename: str, directory: str|None = None) -> None:
        """
        Save the figure as a SVG (Simple Vector Graphic) file with the specified resolution.
        The resolution for the file may be set differently to the display resolution using `resolution_for_files`.

        Parameters:
            str filename:
                The name of the file to save the figure as (without extension).
            str|None directory:
                The directory to save the file in. If None, the current working directory is used.
        """
        filepath = f"{filename}.svg"
        if directory is not None:
            filepath = os.path.join(directory, filepath)
        if not self.__locked:
            raise RuntimeError("Figure has not been created yet. Call `make_figure_and_axes` first.")
        self.__figure.savefig(filepath, dpi = self.resolution_for_files if self.resolution_for_files is not None else self.resolution)
